/*
 * Copyright (C) 2005-2006 Andre Noll <maan@systemlinux.org>
 *
 *     This program is free software; you can redistribute it and/or modify
 *     it under the terms of the GNU General Public License as published by
 *     the Free Software Foundation; either version 2 of the License, or
 *     (at your option) any later version.
 *
 *     This program is distributed in the hope that it will be useful,
 *     but WITHOUT ANY WARRANTY; without even the implied warranty of
 *     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *     GNU General Public License for more details.
 *
 *     You should have received a copy of the GNU General Public License
 *     along with this program; if not, write to the Free Software
 *     Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111, USA.
 */

/** \file alsa_writer.c paraslash's alsa output plugin */

/*
 * Based in parts on aplay.c from the alsa-utils-1.0.8 package,
 * Copyright (c) by Jaroslav Kysela <perex@suse.cz>, which is
 * based on the vplay program by Michael Beck.
 */

#include "para.h"
#include "fd.h"
#include "string.h"
#include "list.h"
#include "sched.h"
#include "write.h"

#include <alsa/asoundlib.h>

#include "alsa_write.cmdline.h"
#include "error.h"


#define FORMAT SND_PCM_FORMAT_S16_LE

/** data specific to the alsa writer */
struct private_alsa_data {
	/** the alsa handle */
	snd_pcm_t *handle;
	/** determined and set by alsa_open() */
	size_t bytes_per_frame;
	/** don't write anything until this time */
	struct timeval next_chunk;
	/** the return value of snd_pcm_hw_params_get_buffer_time_max() */
	unsigned buffer_time;
	unsigned samplerate;
	unsigned channels;
};

/*
 * open and prepare the PCM handle for writing
 *
 * Install PCM software and hardware configuration. Exit on errors.
 */
static int alsa_open(struct writer_node *w)
{
	snd_pcm_hw_params_t *hwparams;
	snd_pcm_sw_params_t *swparams;
	snd_pcm_uframes_t buffer_size, xfer_align, start_threshold,
		stop_threshold;
	int err;
	snd_pcm_info_t *info;
	snd_pcm_uframes_t period_size;
	struct private_alsa_data *pad = para_calloc(sizeof(struct private_alsa_data));
	struct alsa_write_args_info *conf = w->conf;
	struct writer_node_group *wng = w->wng;

	if (!conf->samplerate_given && wng->samplerate)
		pad->samplerate = *wng->samplerate;
	else
		pad->samplerate = conf->samplerate_arg;
	if (!conf->channels_given && wng->channels)
		pad->channels = *wng->channels;
	else
		pad->channels = conf->channels_arg;
	PARA_INFO_LOG("%d channel(s), %dHz\n", pad->channels, pad->samplerate);
	w->private_data = pad;
	snd_pcm_info_alloca(&info);
	err = snd_pcm_open(&pad->handle, conf->device_arg,
		SND_PCM_STREAM_PLAYBACK, SND_PCM_NONBLOCK);
	if (err < 0)
		return -E_PCM_OPEN;
	if ((err = snd_pcm_info(pad->handle, info)) < 0)
		return -E_SND_PCM_INFO;

	snd_pcm_hw_params_alloca(&hwparams);
	snd_pcm_sw_params_alloca(&swparams);
	if (snd_pcm_hw_params_any(pad->handle, hwparams) < 0)
		return -E_BROKEN_CONF;
	if (snd_pcm_hw_params_set_access(pad->handle, hwparams,
			SND_PCM_ACCESS_RW_INTERLEAVED) < 0)
		return -E_ACCESS_TYPE;
	if (snd_pcm_hw_params_set_format(pad->handle, hwparams, FORMAT) < 0)
		return -E_SAMPLE_FORMAT;
	if (snd_pcm_hw_params_set_channels(pad->handle, hwparams,
			pad->channels) < 0)
		return -E_CHANNEL_COUNT;
	if (snd_pcm_hw_params_set_rate_near(pad->handle, hwparams,
			&pad->samplerate, 0) < 0)
		return -E_SET_RATE;
	err = snd_pcm_hw_params_get_buffer_time_max(hwparams, &pad->buffer_time, 0);
	if (err < 0 || !pad->buffer_time)
		return -E_GET_BUFFER_TIME;
	PARA_INFO_LOG("buffer time: %d\n", pad->buffer_time);
	if (snd_pcm_hw_params_set_buffer_time_near(pad->handle, hwparams,
			&pad->buffer_time, 0) < 0)
		return -E_SET_BUFFER_TIME;
	if (snd_pcm_hw_params(pad->handle, hwparams) < 0)
		return -E_HW_PARAMS;
	snd_pcm_hw_params_get_period_size(hwparams, &period_size, 0);
	snd_pcm_hw_params_get_buffer_size(hwparams, &buffer_size);
	PARA_INFO_LOG("buffer size: %lu, period_size: %lu\n", buffer_size,
		period_size);
	if (period_size == buffer_size)
		return -E_BAD_PERIOD;
	snd_pcm_sw_params_current(pad->handle, swparams);
	err = snd_pcm_sw_params_get_xfer_align(swparams, &xfer_align);
	if (err < 0 || !xfer_align)
		return -E_GET_XFER;
	snd_pcm_sw_params_set_avail_min(pad->handle, swparams, period_size);
	/* round to closest transfer boundary */
	start_threshold = (buffer_size / xfer_align) * xfer_align;
	if (start_threshold < 1)
		start_threshold = 1;
	if (snd_pcm_sw_params_set_start_threshold(pad->handle, swparams,
			start_threshold) < 0)
		return -E_START_THRESHOLD;
	stop_threshold = buffer_size;
	if (snd_pcm_sw_params_set_stop_threshold(pad->handle, swparams,
			stop_threshold) < 0)
		return -E_STOP_THRESHOLD;
	if (snd_pcm_sw_params_set_xfer_align(pad->handle, swparams,
			xfer_align) < 0)
		return -E_SET_XFER;
	if (snd_pcm_sw_params(pad->handle, swparams) < 0)
		return -E_SW_PARAMS;
	pad->bytes_per_frame = snd_pcm_format_physical_width(FORMAT)
		* pad->channels / 8;
	if (snd_pcm_nonblock(pad->handle, 1))
		PARA_ERROR_LOG("%s\n", "failed to set nonblock mode");
	return period_size * pad->bytes_per_frame;
}

static int alsa_write_pre_select(struct sched *s, struct writer_node *wn)
{
	struct private_alsa_data *pad = wn->private_data;
	struct writer_node_group *wng = wn->wng;
	struct timeval diff;

	if (*wng->loaded < pad->bytes_per_frame)
		return 1;
	if (tv_diff(now, &pad->next_chunk, &diff) < 0) {
		if (tv_diff(&s->timeout, &diff, NULL) > 0)
			s->timeout = diff;
	} else {
		s->timeout.tv_sec = 0;
		s->timeout.tv_usec = 1;
	}
	return 1;
//	PARA_INFO_LOG("timeout: %lu\n", tv2ms(&s->timeout));
}

static int alsa_write_post_select(__a_unused struct sched *s,
		struct writer_node *wn)
{
	struct private_alsa_data *pad = wn->private_data;
	struct writer_node_group *wng = wn->wng;
	size_t frames = (*wng->loaded - wn->written) / pad->bytes_per_frame;
	snd_pcm_sframes_t ret;
	unsigned char *data = (unsigned char*)wng->buf + wn->written;
	struct timeval tv;

//	PARA_INFO_LOG("%zd frames\n", frames);
	if (!frames) {
		if (*wng->input_eof)
			wn->written = *wng->loaded;
		return 1;
	}
	if (tv_diff(now, &pad->next_chunk, NULL) < 0)
		return 1;
	ret = snd_pcm_writei(pad->handle, data, frames);
	if (ret == -EPIPE) {
		PARA_WARNING_LOG("%s", "EPIPE\n");
		snd_pcm_prepare(pad->handle);
		return 1;
	}
	if (ret < 0) {
		PARA_WARNING_LOG("%s", "ALSA ERROR\n");
		return -E_ALSA_WRITE;
	}
	ms2tv(pad->buffer_time / 4000, &tv);
//	ms2tv(1, &tv);
	tv_add(now, &tv, &pad->next_chunk);
	wn->written += ret * pad->bytes_per_frame;
	return 1;
}

static void alsa_close(struct writer_node *wn)
{
	struct private_alsa_data *pad = wn->private_data;
	PARA_INFO_LOG("closing writer node %p\n", wn);
	snd_pcm_drain(pad->handle);
	snd_pcm_close(pad->handle);
	snd_config_update_free_global();
	free(pad);
}

__malloc void *alsa_parse_config(char *options)
{
	struct alsa_write_args_info *conf
		= para_calloc(sizeof(struct alsa_write_args_info));
	PARA_INFO_LOG("options: %s, %d\n", options, strcspn(options, " \t"));
	int ret = alsa_cmdline_parser_string(options, conf, "alsa_write");
	if (ret)
		goto err_out;
	PARA_INFO_LOG("help given: %d\n", conf->help_given);
	return conf;
err_out:
	free(conf);
	return NULL;
}

/** the init function of the alsa writer */
void alsa_writer_init(struct writer *w)
{
	w->open = alsa_open;
	w->close = alsa_close;
	w->pre_select = alsa_write_pre_select;
	w->post_select = alsa_write_post_select;
	w->parse_config = alsa_parse_config;
	w->shutdown = NULL; /* nothing to do */
}
