/*
 * Copyright (C) 1997-2006 Andre Noll <maan@systemlinux.org>
 *
 *     This program is free software; you can redistribute it and/or modify
 *     it under the terms of the GNU General Public License as published by
 *     the Free Software Foundation; either version 2 of the License, or
 *     (at your option) any later version.
 *
 *     This program is distributed in the hope that it will be useful,
 *     but WITHOUT ANY WARRANTY; without even the implied warranty of
 *     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *     GNU General Public License for more details.
 *
 *     You should have received a copy of the GNU General Public License
 *     along with this program; if not, write to the Free Software
 *     Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111, USA.
 */

/** \file client.c the client program used to connect to para_server */

#include "para.h"
#include "list.h"
#include "sched.h"
#include "client.cmdline.h"
#include "string.h"
#include "stdin.h"
#include "stdout.h"
#include "client.h"
#include "error.h"

INIT_CLIENT_ERRLISTS;

static struct private_client_data *pcd;
static struct stdin_task sit;
static struct stdout_task sot;

/*
 * client log function
 */
void para_log(int ll, const char* fmt,...)
{
	va_list argp;

	/* ignore log message if loglevel is not high enough */
	if (pcd && ll < pcd->conf.loglevel_arg)
		return;
	va_start(argp, fmt);
	vfprintf(stderr, fmt, argp);
	va_end(argp);
}

static void client_event_handler(struct task *t)
{
	struct private_client_data *p = t->private_data;

	PARA_NOTICE_LOG("%s\n", PARA_STRERROR(-t->ret));
	if (t->ret != -E_HANDSHAKE_COMPLETE) {
		unregister_task(t);
		p->eof = 1;
		return;
	}
	if (p->status == CL_SENDING) {
		stdin_set_defaults(&sit);
		sit.buf = para_malloc(sit.bufsize),
		register_task(&sit.task);
		p->inbuf = sit.buf;
		p->in_loaded = &sit.loaded;
		p->in_eof = &sit.eof;
		return;
	}
	stdout_set_defaults(&sot);
	sot.buf = p->buf;
	sot.loaded = &p->loaded;
	sot.input_eof = &p->eof;
	register_task(&sot.task);
}

/*
 * MAIN
 */
int main(int argc, char *argv[])
{

	int ret;
	struct sched s;

	s.default_timeout.tv_sec = 1;
	s.default_timeout.tv_usec = 0;
	ret = client_parse_config(argc, argv, &pcd);
	if (ret < 0)
		goto out;
	pcd->task.event_handler = client_event_handler;
	ret = client_open(pcd);
	if (ret < 0)
		goto out;
	ret = sched(&s);
	client_close(pcd);
out:
	if (ret < 0)
		PARA_ERROR_LOG("%s\n", PARA_STRERROR(-ret));
	return ret >= 0? EXIT_SUCCESS: EXIT_FAILURE;
}
